#!/bin/bash

# ---------------------------------------------------------------------------- #
# This file is part of gsbuild
# Copyright (C) 2005 Greg Schafer <gschafer@zip.com.au>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Name:      gsbuild_cross.sh
# Synopsis:  A sample implementation of the "DIY Linux Cross Compilation Howto"
# ---------------------------------------------------------------------------- #

# The guts of the host_tools phase.
do_host_tools() {
	# Declare our variables needed for this phase. Export PATH only.
	SCRATCH_DIR=$ORIG_SCRATCH_DIR/host_tools
	local PHASE=host_tools
	local PKG_SCRIPTS_DIR="${CURDIR}/scriptlets/${PHASE}"
	local LOGS_DIR="${SCRATCH_DIR}/logs/${PHASE}"
	local STAMP_DIR="${SCRATCH_DIR}/stamp/${PHASE}"
	local TIME_REPORT="${LOGS_DIR}/TIME.REPORT"
	export PATH="${HOST_TOOLS_PFX}/bin:/bin:/usr/bin"
	local PKG

	# We need these directories.
	test -d "$LOGS_DIR" || mkdir -p $LOGS_DIR
	test -d "$STAMP_DIR" || mkdir -p $STAMP_DIR

	if [ ! -r "${STAMP_DIR}/${PHASE}" ]; then
		# Set up the time report.
		time_report start

		# Build the packages.
		for PKG in $HOST_TOOLS_PKGS; do
			do_build $PKG
		done

		# Finish up the time report.
		time_report end

		echo -e "\n${GREEN}SUCCESS:${OFF} Build of Host Tools Phase Finished!\n"
		do_stamp $PHASE
	else
		echo -e "\n${CYAN}NOTICE:${OFF} Host Tools Phase already completed, skipping.\n"
	fi
}

# The guts of the cross_tools phase.
do_cross_tools() {
	# Declare our variables needed for this phase. Export PATH only.
	SCRATCH_DIR=$ORIG_SCRATCH_DIR/$BASE_TARGET/$THREAD_LIB
	local PHASE=cross_tools
	local PKG_SCRIPTS_DIR="${CURDIR}/scriptlets/${PHASE}"
	local LOGS_DIR="${SCRATCH_DIR}/logs/${PHASE}"
	local STAMP_DIR="${SCRATCH_DIR}/stamp/${PHASE}"
	local TIME_REPORT="${LOGS_DIR}/TIME.REPORT"
	export PATH="${CROSS_TOOLS_PFX}/bin:${HOST_TOOLS_PFX}/bin:/bin:/usr/bin"
	local PKG

	# We need these directories.
	test -d "$LOGS_DIR" || mkdir -p $LOGS_DIR
	test -d "$STAMP_DIR" || mkdir -p $STAMP_DIR

	if [ ! -r "${STAMP_DIR}/${PHASE}" ]; then
		# Set up the time report.
		time_report start

		# Build the packages.
		for PKG in $CROSS_TOOLS_PKGS; do
			do_build $PKG
		done

		# Finish up the time report.
		time_report end

		echo -e "\n${GREEN}SUCCESS:${OFF} Build of Cross Tools Phase for $TARGET Finished!\n"
		do_stamp $PHASE
	else
		echo -e "\n${CYAN}NOTICE:${OFF} Cross Tools Phase for $TARGET already completed, skipping.\n"
	fi
}

# Simple function to check for presence of essential config files/dirs.
check_for_conf_file() {
	test -r "$1" || {
		echo -e "\n\e[1;31mERROR:\e[0m Cannot find requisite \"$1\" $2!\n" >&2
		exit 1
	}
}

# Simple function to show options.
show_options() {
	echo -e "\nOptions Summary:\n"
	echo -e "\tTemptools Style Build:\t`test "$TEMPTOOLS_STYLE" = "Y" && echo Yes || echo No`"
}

# Simple function to clean up when exiting via our exit trap.
clean_up () {
	rm -f /tmp/*.$$ >/dev/null 2>&1
}

# ---------------------------------------------------------------------------- #
# Main body of script.
# ---------------------------------------------------------------------------- #

# We rely on features found only in bash-2.x. Seeing as we developed and tested
# with 2.05b, just require it as the minimum.
case "$BASH_VERSION" in
2.05[b-z]* | [3-9].*) ;;
*)	echo >&2
	echo "ERROR: Need bash version 2.05b or greater!" >&2
	echo "       Please see docs for info on how to proceed." >&2
	echo >&2
	exit 1 ;;
esac

# Parse the command line args.
DRY_RUN=""
while [ $# -gt 0 ]; do
	case "$1" in
	--dry-run)	DRY_RUN=1 ;;
	*)		echo -e "\nUsage: ${0##*/} [switches]\n" >&2
			echo -e "  --dry-run\t: perform pre-build sanity checks" >&2
			exit 1 ;;
	esac
	shift
done

# Bomb on simple errors.
# Important - our whole error trapping strategy relies on this. Note - the ERR
# trap is lost inside our while loops as per bash docs.
set -e
trap 'echo -e "\n\e[1;31mERROR:\e[0m Simple error trapped!\n" >&2' ERR

# Clean up on exit. Handle received signals separately.
trap 'clean_up' 0
trap 'exit 2' 1 2 3 15

# Be where we wanna be. This is essential for when relauching ourself upon
# entering the chroot.
if [[ $0 == */* ]]; then
	cd ${0%/*}/
fi

# Initialize some variables.
BUILD=cross
CURDIR=`pwd`
PKG_LIST=package-base.list
TMP_FILE=/tmp/tmp.$$

# Switch off hashing. Set a sane umask.
set +h
umask 022

# Check for the 2 main config files, if found, source them. Verify all the user
# supplied environment variables. Ensure scriptlets have been copied over.
check_for_conf_file build.conf.cross file && . ./build.conf.cross
check_for_conf_file common-sh.functions file && . ./common-sh.functions
verify_vars "TARGET THREAD_LIB BASE_TARGET BUILD_HOST CONFIG_SITE \
  CROSS_PFX CROSS_TOOLS_PFX GLIBC_HEADERS HOST_TOOLS_PFX \
  LC_ALL LDFLAGS PATH SYSROOT TGT_PFX TT_PFX TZ"
verify_vars "SCRATCH_DIR TARBALLS_DIR PATCHES_DIR TEMPTOOLS_STYLE" conf
verify_yn_vars "TEMPTOOLS_STYLE"
check_for_conf_file scriptlets/host_tools dir
check_for_conf_file scriptlets/cross_tools dir
check_for_conf_file $PKG_LIST file
check_for_conf_file config.sub file

# Set up the package lists.
HOST_TOOLS_PKGS="make sed binutils gcc"
CROSS_TOOLS_PKGS="linux-libc-headers glibc-pass0 binutils gcc-pass1 glibc gcc-pass2"
test "$BASE_TARGET" = ia64 && GLIBC_HEADERS=Y
test "$GLIBC_HEADERS" = N && \
	CROSS_TOOLS_PKGS=`echo $CROSS_TOOLS_PKGS | sed "s/glibc-pass0//"`
REAL_PKGS=`
	for P in $HOST_TOOLS_PKGS $CROSS_TOOLS_PKGS; do
		if echo $P | grep "\-pass.$" 2>&1 >/dev/null; then
			P=${P%-*}
		fi
		echo $P 
	done | sort | uniq`

# Now that we have our functions and data, perform some more sanity checks.
verify_dirs_exist "SCRATCH_DIR TARBALLS_DIR PATCHES_DIR CROSS_PFX"
verify_dir_writable "SCRATCH_DIR CROSS_PFX"
verify_pkg_list
verify_scriptlets host_tools HOST_TOOLS_PKGS
verify_scriptlets cross_tools CROSS_TOOLS_PKGS
verify_pkg_list_fields
verify_tarballs
verify_patches host_tools HOST_TOOLS_PKGS
verify_patches cross_tools CROSS_TOOLS_PKGS

# Sort out the host and target variables.
if ! ./config.sub $TARGET >/dev/null 2>&1; then
	echo -e "\n${RED}ERROR:${OFF} $TARGET is an invalid target!" >&2
	echo -e "       ie: it's not recognized by config.sub\n" >&2
	exit 1
fi

CANON_TARGET=`./config.sub $TARGET`

if ! echo $CANON_TARGET | grep linux >/dev/null; then
	echo -e "\n${RED}ERROR:${OFF} $CANON_TARGET is an unsupported target!" >&2
	echo -e "       ie: it's not linux\n" >&2
	exit 1
fi

# Fake the host if necessary.
if [ "$BUILD_HOST" = "$CANON_TARGET" ]; then
	VENDOR=`echo $BUILD_HOST | awk -F - '{ print $2 }'`
	BUILD_HOST=${BUILD_HOST/$VENDOR/fake}
fi

# Extra sanity check
case "$TARGET" in
alpha* | arm* | cris* | hppa* | i?86* | ia64* | \
m68k* | mips* | powerpc64* | powerpc* | ppc* | \
s390* | sh* | sparc64* | sparc* | x86_64*) ;;
*)	echo -e "\n${RED}ERROR:${OFF} $TARGET is an unsupported target!" >&2
	echo -e "       ie: no kernel headers available\n" >&2
	exit 1 ;;
esac

# We cannot be superuser in this phase. It's far too dangerous.. unless you're a
# real bloody idiot.
if [ "`id -u`" = "0" -a "$BLOODYIDIOT" != YES ]; then
	echo -e "\n${RED}ERROR:${OFF} Cannot build Host Tools Phase as root!\n" >&2
	exit 1
fi

# Handle the "--dry-run" argument.
test "$DRY_RUN" && {
	show_options
	echo -e "\n${GREEN}SUCCESS:${OFF} All pre-build sanity checks passed!\n"
	exit 0
}

# Do the business.
BINUTILS_VER=`get_pkg_field binutils 2`
GCC_VER=`get_pkg_field gcc 2`
LLH_VER=`get_pkg_field linux-libc-headers 2`
GLIBC_VER=`get_pkg_field glibc 2`
ORIG_SCRATCH_DIR=$SCRATCH_DIR
test "$TEMPTOOLS_STYLE" = Y && TGT_PFX=$TT_PFX

do_host_tools
do_cross_tools

exit 0
