#!/temptools/bin/bash
#
# *** SEE DOCS FOR HOW TO RUN SCRIPT - ABOVE INTERPRETER PATH IS NOT A TYPO! ***
#
# ---------------------------------------------------------------------------- #
# This file is part of gsbuild
# Copyright (C) 2004, 2005 Greg Schafer <gschafer@zip.com.au>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Name:      gsbuild.sh
# Synopsis:  A sample implementation of the "DIY Linux Reference Build" (x86)
# Features:  * Optional Iterative Comparison Analysis (ICA)
#            * Optional Package Manager integration
#            * Is restartable ie: uses stamp files to remember where it's up to
# ---------------------------------------------------------------------------- #

# The guts of the temptools phase.
do_temptools() {
	# Declare our variables needed for this phase. Export PATH only.
	local PKG_SCRIPTS_DIR="${CURDIR}/scriptlets/${PHASE}"
	local LOGS_DIR="${SCRATCH_DIR}/logs/${PHASE}"
	local STAMP_DIR="${SCRATCH_DIR}/stamp/${PHASE}"
	local TIME_REPORT="${LOGS_DIR}/TIME.REPORT"
	export PATH="${TT_PFX}/bin:/bin:/usr/bin"
	local PKG

	# We need these directories.
	test -d "$LOGS_DIR" || mkdir -p $LOGS_DIR
	test -d "$STAMP_DIR" || mkdir -p $STAMP_DIR

	# Set up the time report.
	time_report start

	# Build the packages.
	for PKG in $TEMPTOOLS_PKGS; do
		do_build $PKG
	done

	# Finish up the time report.
	time_report end

	echo -e "\n${GREEN}SUCCESS:${OFF} Build of Temptools Phase Finished!\n"
}

# The guts of the chroot phase.
do_chroot() {
	# Declare some of our variables needed for this phase.
	local PKG_SCRIPTS_DIR="${CURDIR}/scriptlets/${PHASE}"
	local ITERATIONS=1
	test "$ICA" = Y && ITERATIONS=3
	local PKG
	local -a ITER1_ONLY=(create_dirs do_mounts create_symlinks create_misc_files
		create_users MAKEDEV linux-libc-headers man-pages readjust-toolchain)

	# Use this variable as the iteration loop counter.
	local ITER=1

	# Ensure the virtual file systems are mounted. If /etc/mtab exists then
	# the real do_mounts was already done and we've likely been restarted.
	test -r /etc/mtab && . ${PKG_SCRIPTS_DIR}/do_mounts

	while [ "$ITER" -le "$ITERATIONS" ]; do

		# Remaining variables needed for this phase. Export PATH only.
		local LOGS_DIR="${SCRATCH_DIR}/logs/${PHASE}-${ITER}"
		local STAMP_DIR="${SCRATCH_DIR}/stamp/${PHASE}-${ITER}"
		local TIME_REPORT="${LOGS_DIR}/TIME.REPORT"

		if [ "$ITER" != 1 ]; then
			export PATH="/bin:/usr/bin:/sbin:/usr/sbin"
		fi

		# We need these directories.
		test -d "$LOGS_DIR" || mkdir -p $LOGS_DIR
		test -d "$STAMP_DIR" || mkdir -p $STAMP_DIR

		# PM Stuff
		if [ "$PKG_MGR" ]; then
			local ARC_DIR="${SCRATCH_DIR}/archives"
			test -d "$ARC_DIR/specfiles" || mkdir -p $ARC_DIR/specfiles
			export ITER RUN_TESTS RUN_OPTIONAL_TESTS
		fi

		# Set up the time report.
		time_report start

		# Start building the packages.
		if [ ! -r "${STAMP_DIR}/../iter${ITER}" ]; then

			for PKG in $CHROOT_PKGS; do
				if [ "$ITER" = 1 ]; then
					do_build $PKG
				elif [ "$ITER" != 1 ] && \
					! echo " ${ITER1_ONLY[*]} " | grep " $PKG " 2>&1 >/dev/null; then
					do_build $PKG
				fi
			done

			# Some final cleanups.
			(cd /usr/share/info; rm -f dir
			for f in *; do install-info $f dir 2>/dev/null || :; done)
			rm -rf /tmp/*
			/sbin/ldconfig
			sed -i '/dummy/d' /etc/{group,passwd}
			/usr/sbin/pwconv
			/usr/sbin/grpconv

			# Do the ICA prep if applicable.
			test "$ICA" = Y && do_ica_prep

			echo -e "\n${GREEN}SUCCESS:${OFF} Build of Chroot Phase Iteration $ITER Finished!\n"

			echo done > ${STAMP_DIR}/../iter${ITER}

		fi

		# We're done. Finish up the time report and bump the counter.
		time_report end
		((ITER += 1))

	done
}

# Prepare the scriptlets and packagedata.
do_prep() {
	check_for_conf_file gsbuild_${BUILD}.prep file
	> $TMP_FILE

	if ! sed -i q $TMP_FILE >/dev/null 2>&1; then
		echo -e "\n${RED}ERROR:${OFF} need sed version 4 or newer! (for the -i option)\n" >&2
		exit 1
	fi

	if [ ! -r "scriptlets/prep-${BUILD}.done" ]; then
		echo -en "\nApplying sed transformations to scriptlets... "
		(. ./gsbuild_${BUILD}.prep) || {
			echo -e "\n${RED}ERROR:${OFF} problem executing gsbuild_${BUILD}.prep!\n" >&2
			exit 1
		}
		echo "done."

		echo -en "Creating $PKG_LIST from scriptlets/packagedata... "
		cat > $PKG_LIST << "EOF"
# ---------------------------------------------------------------------------- #
# gsbuild configuration file - Tailor as needed.
#
# The format is a simple TAB separated list of fields forming 5 columns. Blank
# lines and lines beginning with `#' are ignored (a-la shell style). Column
# contents should be self-explanatory. NOTE - all 5 fields must be occupied!
# ---------------------------------------------------------------------------- #
EOF
		cat scriptlets/packagedata >> $PKG_LIST
		echo "done."

		echo done > scriptlets/prep-${BUILD}.done
	else
		echo -e "\n${CYAN}NOTICE:${OFF} script preparation already done.\n"
		exit 0
	fi
}

# Simple function to check for presence of essential config files/dirs.
check_for_conf_file() {
	test -r "$1" || {
		echo -e "\n\e[1;31mERROR:\e[0m Cannot find requisite \"$1\" $2!\n" >&2
		exit 1
	}
}

# Simple function to show options.
show_options() {
	echo -e "\nOptions Summary:\n"
	echo -e "\tNormal Test Suites:\t`test "$RUN_TESTS" = "Y" && echo Yes || echo No`"
	echo -e "\tOptional Test Suites:\t`test "$RUN_OPTIONAL_TESTS" = "Y" && echo Yes || echo No`"
	echo -e "\tICA Build:\t\t`test "$ICA" = "Y" && echo Yes || echo No`"
	echo -e "\tPM Build:\t\t`test "$PKG_MGR" && echo "Yes ($PKG_MGR)" || echo No`"
}

# Simple function to clean up when exiting via our exit trap.
clean_up () {
	rm -f /tmp/*.$$ >/dev/null 2>&1
	if [ "$PHASE" = chroot ]; then
		umount /proc /dev/{pts,shm} /sys 2>/dev/null
	fi
}

# ---------------------------------------------------------------------------- #
# Main body of script.
# ---------------------------------------------------------------------------- #

# We rely on features found only in bash-2.x. Seeing as we developed and tested
# with 2.05b, just require it as the minimum.
case "$BASH_VERSION" in
2.05[b-z]* | [3-9].*) ;;
*)	echo >&2
	echo "ERROR: Need bash version 2.05b or greater!" >&2
	echo "       Please see docs for info on how to proceed." >&2
	echo >&2
	exit 1 ;;
esac

# Parse the command line args.
PREP=""
DRY_RUN=""
PHASE=""
while [ $# -gt 0 ]; do
	case "$1" in
	--prep)		PREP=1 ;;
	--dry-run)	DRY_RUN=1 ;;
	--chroot)	PHASE=chroot ;;
	*)		echo -e "\nUsage: ${0##*/} [switches]\n" >&2
			echo -e "  --prep\t: prepare the scriptlets" >&2
			echo -e "  --dry-run\t: perform pre-build sanity checks" >&2
			echo -e "  --chroot\t: build target system inside chroot environment\n" >&2
			exit 1 ;;
	esac
	shift
done

# Bomb on simple errors.
# Important - our whole error trapping strategy relies on this. Note - the ERR
# trap is lost inside our while loops as per bash docs.
set -e
trap 'echo -e "\n\e[1;31mERROR:\e[0m Simple error trapped!\n" >&2' ERR

# Clean up on exit. Handle received signals separately.
trap 'clean_up' 0
trap 'exit 2' 1 2 3 15

# Be where we wanna be. This is essential for when relauching ourself upon
# entering the chroot.
if [[ $0 == */* ]]; then
	cd ${0%/*}/
fi

# Initialize some variables.
BUILD=base
CURDIR=`pwd`
: ${PHASE:=temptools}
PKG_LIST=package-${BUILD}.list
TMP_FILE=/tmp/tmp.$$

# Switch off hashing. Set a sane umask.
set +h
umask 022

# Must have `/dev/null' early inside chroot else bad things will happen.
if [ "$PHASE" = chroot ]; then
	test -d /dev || mkdir /dev
	test -c /dev/null || { rm -f /dev/null; mknod /dev/null c 1 3; }
fi

# Check for the 2 main config files, if found, source them. Verify all the user
# supplied environment variables. Ensure scriptlets have been copied over.
check_for_conf_file build.conf file && . ./build.conf
check_for_conf_file common-sh.functions file && . ./common-sh.functions
verify_vars "CONFIG_SITE LC_ALL LDFLAGS TT_PFX TZ"
test "$PHASE" = temptools && verify_vars "SYSROOT"
verify_vars "SCRATCH_DIR TARBALLS_DIR PATCHES_DIR ICA RUN_TESTS RUN_OPTIONAL_TESTS" conf
verify_yn_vars "ICA RUN_TESTS RUN_OPTIONAL_TESTS"
verify_pm
check_for_conf_file scriptlets/temptools dir
check_for_conf_file scriptlets/chroot dir

# Handle the "--prep" argument.
test "$PREP" && {
	do_prep
	echo -e "\n${GREEN}SUCCESS:${OFF} script preparation complete!\n"
	exit 0
}

# Sanity check the prep.
verify_prep
check_for_conf_file $PKG_LIST file

# On some systems 'pwd' returns the "real" cwd instead of what we need e.g:
# "/mnt/sysroot/temptools/src/gsbuild" instead of "/temptools/src/gsbuild".
# Strip off any occurrence of the unwanted prefix.
CURDIR=${CURDIR#${SYSROOT}}

# Set up the package lists.
MISC_PKGS=(adjust-toolchain remove_cruft create_dirs do_mounts create_symlinks
	create_misc_files create_users readjust-toolchain MAKEDEV bpm)

PACMAN_PKGS="zlib libtar pacman"
BPM_PKGS="ogdlutils cpio hashsum bpm"

case "$PKG_MGR" in
PACMAN)	PM_PKGS=$PACMAN_PKGS ;;
BPM)	PM_PKGS=$BPM_PKGS ;;
*)	PM_PKGS="" ;;
esac

TEMPTOOLS_PKGS="
  make-pass1 sed-pass1 binutils-pass1 gcc-pass1 linux-libc-headers glibc adjust-toolchain
  gcc-pass2 binutils-pass2 gawk coreutils bzip2 gzip diffutils findutils make-pass2 grep
  sed-pass2 gettext ncurses patch tar texinfo bash-pass2 m4 bison flex util-linux perl tcl
  expect dejagnu fakeroot $PM_PKGS remove_cruft"

CHROOT_PKGS="
  create_dirs do_mounts create_symlinks create_misc_files create_users MAKEDEV
  linux-libc-headers man-pages glibc readjust-toolchain gcc binutils coreutils zlib
  debianutils iana-etc findutils gawk ncurses readline vim m4 bison less groff sed flex
  gettext net-tools inetutils procps perl texinfo autoconf automake bash file libtool
  bzip2 diffutils ed kbd e2fsprogs grep grub gzip man make module-init-tools patch psmisc
  shadow sysklogd sysvinit tar util-linux"

REAL_PKGS=`
	for P in $TEMPTOOLS_PKGS $CHROOT_PKGS; do
		if echo $P | grep "\-pass.$" 2>&1 >/dev/null; then
			P=${P%-*}
		fi
		echo $P 
	done | sort | uniq`

for P in ${MISC_PKGS[*]}; do
	REAL_PKGS=`echo $REAL_PKGS | sed "s/$P//"`
done

REAL_PKGS="$REAL_PKGS MAKEDEV"
test "$PKG_MGR" = BPM && REAL_PKGS="$REAL_PKGS bpm"

# Now that we have our functions and data, perform some more sanity checks.
verify_dirs_exist "SCRATCH_DIR TARBALLS_DIR PATCHES_DIR TT_PFX"
test "$PHASE" = temptools && verify_dirs_exist "SYSROOT"
verify_dir_writable "SCRATCH_DIR TT_PFX"
verify_dirs "SCRATCH_DIR TARBALLS_DIR PATCHES_DIR"
verify_phase
verify_pkg_list
verify_scriptlets temptools TEMPTOOLS_PKGS
verify_scriptlets chroot CHROOT_PKGS
verify_pkg_list_fields
verify_tarballs
verify_patches temptools TEMPTOOLS_PKGS
verify_patches chroot CHROOT_PKGS

# We cannot be superuser in the temptools phase. It's far too dangerous.. unless
# you're a real bloody idiot.
if [ "$PHASE" = temptools -a "`id -u`" = "0" -a "$BLOODYIDIOT" != YES ]; then
	echo -e "\n${RED}ERROR:${OFF} Cannot build Temptools Phase as root!\n" >&2
	exit 1
fi

# If performing an ICA, we don't want to run the test suites. Override. Also, we
# don't yet support ICA with PM. Override that too.
if [ "$ICA" = Y ]; then
	RUN_TESTS=N
	RUN_OPTIONAL_TESTS=N
	unset PKG_MGR
fi

# Handle the "--dry-run" argument.
test "$DRY_RUN" && {
	show_options
	echo -e "\n${GREEN}SUCCESS:${OFF} All pre-build sanity checks passed!\n"
	exit 0
}

# Do the business.
if [ "$PHASE" = temptools ]; then
	do_temptools

	CHROOT_CMD="chroot $SYSROOT \
	${TT_PFX}/bin/env -i \
	HOME=/root \
	PATH=/bin:/usr/bin:/sbin:/usr/sbin:${TT_PFX}/bin \
	TERM=${TERM} \
	CONFIG_SITE=/etc/config.site \
	LC_ALL=${LC_ALL} \
	LDFLAGS=\"${LDFLAGS}\" \
	${MAKEFLAGS+\"MAKEFLAGS=$MAKEFLAGS\"} \
	TT_PFX=${TT_PFX} \
	TZ=${TZ} \
	${CURDIR}/${0##*/} --chroot"

	echo -e "${CYAN}NOTICE:${OFF} Please supply root password to enter the chroot environment"
	echo -e "        or press Ctrl-C to exit gracefully.\n"
	test "`id -u`" = "0" -a "$BLOODYIDIOT" != YES && exec "$CHROOT_CMD" || exec su -c "$CHROOT_CMD"
else
	do_chroot

	# Build the ICA reports if applicable.
	if [ "$ICA" = Y ]; then
		if [ ! -r "${SCRATCH_DIR}/stamp/ica" ]; then
			echo -e "${BORDER}\n${CYAN}[ICA] - starting ICA Reports Generation.${OFF}\n"

			do_ica_work 1 2
			do_ica_work 2 3

			echo -e "\n${CYAN}[ICA] - ICA Reports Generation complete.${OFF}\n${BORDER}"
			echo -e "\n${GREEN}SUCCESS:${OFF} Chapter 6 Full ICA Build Finished!\n"
			echo done > ${SCRATCH_DIR}/stamp/ica
		fi
	fi
fi

exit 0
